# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.htmlfrom setuptools import setup, find_packages
# meant to define package metadata in one and only one place

__title__ = 'skater'
__version__ = '1.0.0b.11'
__summary__ = 'Model Agnostic Interpretation Library'
__uri__ = 'https://github.com/datascienceinc/skater/'
__author__ = 'Aaron Kramer, DataScience.com Team, Pramit Choudhary'
__email__ = 'dstm@datascience.com'
__license__ = 'MIT'
__docs__ = 'https://datascienceinc.github.io/Skater/'
