"""Base method for all global interpretations. Is
a subclass of base ModelInterpreter"""

from ..model_interpreter import ModelInterpreter


class BaseGlobalInterpretation(ModelInterpreter):
    """Base class for global model interpretations"""
    pass
